# -*- coding: utf-8 -*-

# foglamp_BEGIN
# See: http://foglamp-iot.readthedocs.io/
# foglamp_END

""" Coin Detector Plugin
"""
__author__ = "Bryce Kroencke"
__copyright__ = "Copyright (c) 2020 Dianomic Systems Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import asyncio
import copy
import subprocess
import logging
import os
import time
import threading
from threading import Thread
import cv2
import numpy as np
import pytz
from datetime import datetime

from sense_hat import SenseHat

from google.cloud import storage
import google.cloud.storage

from foglamp.common import logger
from foglamp.plugins.common import utils
import async_ingest

from foglamp.plugins.south.coin_detection.videostream import VideoStream
from foglamp.plugins.south.coin_detection.inference import Inference
from foglamp.plugins.south.coin_detection.web_stream import WebStream

import tflite_runtime.interpreter as tflite

EDGETPU_SHARED_LIB = 'libedgetpu.so.1'



_LOGGER = logger.setup(__name__, level=logging.DEBUG)

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Coin Detection on FogLAMP',
        'type': 'string',
        'default': 'coin_detection',
        'readonly': 'true'
    },
    'model_file': {
        'description': 'TFlite model file to use for inference',
        'type': 'string',
        'default': 'coin_model.tflite',
        'order': '1',
        'displayName': 'TFlite Model File'
    },
    'labels_file': {
        'description': 'Labels file used during inference',
        'type': 'string',
        'default': 'coin_labels.txt',
        'order': '2',
        'displayName': 'Labels File'
    },
    'gcp_credentials_json': {
        'description': 'JSON file containing credentials for sending to GCP bucket',
        'type': 'string',
        'default': '/home/pi/Downloads/gcp_creds.json',
        'order': '3',
        'displayName': 'GCP Credentials JSON'
    },
    'gcp_bucket': {
        'description': 'Name of GCP bucket',
        'type': 'string',
        'default': 'foglamp-coin-detection',
        'order': '3',
        'displayName': 'GCP Bucket Name'
    },
    'mode': {
        'description': 'Run the plugin in collect mode or in predict mode',
        'type': 'enumeration',
        'default': 'collect',
        'order': '3',
        'options': ['collect', 'predict'],
        'displayName': 'Mode'
    },
    'gcp_model_path': {
        'description': 'Path to pull ML model from within GCP bucket',
        'type': 'string',
        'default': 'models/coin_model.tflite',
        'order': '3',
        'displayName': 'GCP Model Path'
    },
    'asset_name': {
        'description': 'Asset ame',
        'type': 'string',
        'default': 'coin_detection',
        'order': '3',
        'displayName': 'Asset Name'
    },
    'enable_edge_tpu': {
        'description': 'Connect Coral Edge TPU and select this option to use TPU for inference',
        'type': 'boolean',
        'default': 'false',
        'order': '4',
        'displayName': 'Enable Edge TPU'
    },
    'min_conf_threshold': {
        'description': 'Threshold to select the detected objects',
        'type': 'float',
        'default': '0.5',
        'minimum': '0',
        'maximum': '1',
        'order': '5',
        'displayName': 'Minimum Confidence Threshold'
    },
    'source': {
        'description': 'Whether take input feed from camera or network stream',
        'type': 'enumeration',
        'default': 'camera',
        'options': ['stream', 'camera'],
        'order': '6',
        'displayName': 'Source of video feed'
    },
    'stream_url': {
        'description': 'The url of the network stream if network stream is to be used.',
        'type': 'string',
        'default': 'rtsp://ip:port',
        'order': '7',
        'validity': "source == \"stream\"",
        'displayName': 'Stream URL'
    },
    'opencv_backend': {
        'description': 'Backend for processing stream from network',
        'type': 'enumeration',
        'default': 'ffmpeg',
        'options': ['ffmpeg'],
        'order': '8',
        'validity': "source == \"stream\"",
        'displayName': 'OpenCV Backend'
    },
    'stream_protocol': {
        'description': 'The protocol being used by streaming server',
        'type': 'enumeration',
        'default': 'udp',
        'options': ['udp'],
        'order': '9',
        'validity': "source == \"stream\"",
        'displayName': 'Stream Protocol'
    },
    'camera_id': {
        'description': 'The number associated with your video device. See /dev in your '
                       'filesystem. Enter 0 to use /dev/video0, and so on.',
        'type': 'integer',
        'default': '0',
        'order': '10',
        'validity': "source == \"camera\"",
        'displayName': 'Camera ID'
    },
    'enable_window': {
        'description': 'Show detection results in a window',
        'type': 'boolean',
        'default': 'false',
        'order': '11',
        'displayName': 'Enable Detection Window'
    },
    'enable_web_streaming': {
        'description': 'Enable web streaming on specified web streaming port',
        'type': 'boolean',
        'default': 'true',
        'order': '12',
        'displayName': 'Enable Web Streaming'
    },
    'web_streaming_port_no': {
        'description': 'Port number for web streaming',
        'type': 'string',
        'default': '8085',
        'order': '13',
        'displayName': 'Web Streaming Port',
        "validity": "enable_web_streaming == \"true\" "
    },
}

# GLOBAL VARIABLES DECLARATION
c_callback = None
c_ingest_ref = None
frame_processor = None
loop = None
async_thread = None
enable_web_streaming = None
web_stream = None
# Keeping a fixed camera resolution for now. Can give it inside configuration. However changing
# it is quite risky because some devices support changing camera resolution through openCV API
# but others simply don't (Like the coral board).
CAMERA_HEIGHT = 480
CAMERA_WIDTH = 640


# GLOBAL GCP VARIABLES
PATH = None
storage_client = None
bucket = None

# GLOBAL SENSEHAT VARIABLES
sense = SenseHat()


X = [255, 0, 0]  # Red
Y = [255, 255, 0] # Yellow
O = [0, 0, 0]  # Off
G = [0, 255, 0] # Green
B = [0, 0, 255] #Blue


LED_BLANK = [
    O, O, Y, Y, Y, Y, O, O,
    O, Y, Y, Y, Y, Y, Y, O,
    Y, Y, Y, Y, Y, Y, Y, Y,
    Y, Y, Y, Y, Y, Y, Y, Y,
    Y, Y, Y, Y, Y, Y, Y, Y,
    Y, Y, Y, Y, Y, Y, Y, Y,
    O, Y, Y, Y, Y, Y, Y, O,
    O, O, Y, Y, Y, Y, O, O
]

LED_HEADS = [
    O, O, Y, Y, Y, Y, O, O,
    O, X, X, X, X, X, X, O,
    Y, X, X, X, X, X, X, Y,
    Y, Y, Y, X, X, Y, Y, Y,
    Y, Y, Y, X, X, Y, Y, Y,
    Y, X, X, X, X, X, X, Y,
    O, X, X, X, X, X, X, O,
    O, O, Y, Y, Y, Y, O, O
]

LED_TAILS = [
    O, O, Y, Y, Y, Y, O, O,
    O, X, X, Y, Y, Y, Y, O,
    Y, X, X, Y, Y, Y, Y, Y,
    Y, X, X, X, X, X, X, Y,
    Y, X, X, X, X, X, X, Y,
    Y, X, X, Y, Y, Y, Y, Y,
    O, X, X, Y, Y, Y, Y, O,
    O, O, Y, Y, Y, Y, O, O
]

LED_LABELING = [
    O, O, O, O, B, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, B, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, O, B, O, O,
    O, O, O, O, B, B, B, O
]

LED_OFF = [
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O
]

LED_NO_COIN = [
    O, O, O, O, B, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, B, O, O, O,
    X, X, X, O, X, O, O, O,
    X, X, X, O, X, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, O, B, O, O,
    O, O, O, O, B, B, B, O
]

LED_MOVE_COIN = [
    O, O, O, O, B, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, B, O, O, O,
    Y, Y, Y, O, Y, O, O, O,
    Y, Y, Y, O, Y, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, O, B, O, O,
    O, O, O, O, B, B, B, O
]

LED_TAILS_RECORDED = [
    O, O, O, O, G, O, O, O,
    O, O, O, O, G, G, G, O,
    O, O, O, O, G, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, O, B, O, O,
    O, O, O, O, B, B, B, O
]

LED_HEADS_RECORDED = [
    O, O, O, O, B, O, O, O,
    O, O, O, O, B, B, B, O,
    O, O, O, O, B, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, O, O, O, O,
    O, O, O, O, G, G, G, O,
    O, O, O, O, O, G, O, O,
    O, O, O, O, G, G, G, O
]


def plugin_info():
    """ Returns information about the plugin.
    Args:
    Returns:
        dict: plugin information
    Raises:
    """

    return {
        'name': 'Coin Detection plugin',
        'version': '2.0.0',
        'mode': 'async',
        'type': 'south',
        'interface': '1.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """ Initialise the plugin with the merged config
        Args:
           config: JSON configuration document for the South plugin configuration category
        Returns:
           data: JSON object to be used in future calls to the plugin
        Raises:
    """
    data = copy.deepcopy(config)
    return data


def plugin_start(handle):
    """ Start the plugin operation - observe video feed from camera device and classify that using the
        configured ML model
        Args:
            handle: handle returned by the plugin initialisation call
        Returns:
            Sends back the reading object (JSON doc) via C/python async ingest interface when it recognizes
            the video frame content
        Raises:
    """
    global frame_processor, loop, async_thread, enable_web_streaming, web_stream

    loop = asyncio.new_event_loop()
    try:

        # some extra config parameters required for the camera_loop function
        # Since pixel value can be from 0 to 255 , so we are considering mean to
        # be (0+255)/2 = 127.5 .
        handle['input_mean'] = 127.5
        handle['input_std'] = 127.5
        handle['camera_height'] = CAMERA_HEIGHT
        handle['camera_width'] = CAMERA_WIDTH


        web_streaming_port_no = int(handle['web_streaming_port_no']['value'])

        if handle['enable_web_streaming']['value'] == 'true':
            enable_web_streaming = True
        else:
            enable_web_streaming = False

        def run():
            global loop
            loop.run_forever()

        global frame_processor
        frame_processor = FrameProcessor(handle)
        _LOGGER.debug(frame_processor.is_camera_functional)
        _LOGGER.debug(frame_processor.interpreter_loaded)
        if frame_processor.is_camera_functional and frame_processor.interpreter_loaded:
            if enable_web_streaming:

                # make shutdown flag of web stream server false.
                WebStream.SHUTDOWN_IN_PROGRESS = False

                web_stream = WebStream(port=web_streaming_port_no).start_web_streaming_server(local_loop=loop)
                async_thread = Thread(target=run, name="Async Thread")
                async_thread.daemon = True
                async_thread.start()

            frame_processor.start()
        else:
            raise Exception("Camera is not functional. Please shutdown and try again. ")

    except Exception as ex:
        _LOGGER.exception("Human detector plugin failed to start. Details: %s", str(ex))
        raise
    else:
        _LOGGER.info("Plugin started")


def check_need_to_shutdown(handle, new_config, parameters_to_check):
    """
    Checks whether shutdown is required if we change configuration.
    Args:
        handle: Old configuration
        new_config: New Configuration
        parameters_to_check: Parameters (list) whose value if changed then shutdown is required

    Returns:
        True or False
    """
    need_to_shutdown = False
    for parameter in parameters_to_check:
        old_value = handle[parameter]['value']
        new_value = new_config[parameter]['value']
        if new_value != old_value:
            need_to_shutdown = True

    return need_to_shutdown


def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin
        it should be called when the configuration of the plugin is changed
        during the operation of the south service.
    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """
    global frame_processor

    parameters_to_check = ['camera_id', 'enable_window', 'enable_web_streaming',
                           'web_streaming_port_no',
                           'source', 'stream_url', 'stream_protocol', 'opencv_backend']
    need_to_shutdown = check_need_to_shutdown(handle, new_config, parameters_to_check)
    if not need_to_shutdown:
        _LOGGER.debug("No need to shutdown")
        frame_processor.handle_new_config(new_config)
        new_handle = plugin_init(new_config)
        return new_handle

    else:

        was_native_window_enabled = handle['enable_window']['value'] == 'true'
        is_native_window_enabled_now = new_config['enable_window']['value'] == 'true'

        # case  was_native_window_enabled  is_native_window_enabled_now
        # 1.          True                       True
        # 2.          True                       False
        # 3.          False                      True
        # 4.          False                      False

        # Only case 4. should be allowed and rest of the cases are causing problems.
        # Refer to the following link for more details
        # https://stackoverflow.com/questions/60737852/opencv-imshow-hangs-if-called-two-times-from-a-thread

        if was_native_window_enabled or is_native_window_enabled_now:
            _LOGGER.warning("Enabling the native window during reconfigure is known to cause problems. "
                            "Restart the plugin manually. ")
            new_handle = plugin_init(new_config)
            return new_handle

        else:
            plugin_shutdown(handle)
            new_handle = plugin_init(new_config)
            plugin_start(new_handle)
            return new_handle


def plugin_shutdown(handle):
    """ Shutdowns the plugin doing required cleanup
        To be called prior to the south service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
        None
    Raises:
        None
    """
    global frame_processor, loop, async_thread, enable_web_streaming, web_stream
    try:

        frame_processor.shutdown_in_progress = True
        # allow the stream to stop
        time.sleep(3)
        # stopping every other thread one by one

        # checking if the thread is started or not.
        if frame_processor.is_camera_functional and frame_processor.interpreter_loaded:
            frame_processor.join()

        frame_processor = None
        if enable_web_streaming:
            if web_stream is not None:
                WebStream.SHUTDOWN_IN_PROGRESS = True
                web_stream.stop_server(loop)
        loop.stop()
        # async_thread.join()
        async_thread = None
        loop = None

        _LOGGER.info('Plugin has shutdown')

    except Exception as ex:
        _LOGGER.exception(str(ex))
        raise


def plugin_register_ingest(handle, callback, ingest_ref):
    """ Required plugin interface component to communicate to South C server

    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C/Python async ingest interface
        ingest_ref: C opaque object required to passed back to C/Python async ingest interface
    """
    _LOGGER.debug("register ingest")
    global c_callback, c_ingest_ref
    c_callback = callback
    c_ingest_ref = ingest_ref


def pull_model():
    _LOGGER.debug("Beginning model pull function")


    local_path = os.path.join("/usr/local/foglamp/python/foglamp/plugins/south/coin_detection/model", "coin_model.tflite")
    remote_path = os.path.join("models", "coin_model.tflite")


    local_exists = os.path.exists(local_path)
    remote_exists = storage.Blob(bucket=bucket, name=remote_path).exists(storage_client)


    if local_exists and remote_exists:
        _LOGGER.debug("Model exists in both local and remote machines")

        # Get last modification date of local tflite model
        # Add UTC localization to allow comparison with GCP timestamp
        local_model_modified = datetime.fromtimestamp(os.path.getmtime(local_path))
        utc=pytz.UTC
        local_model_modified = utc.localize(local_model_modified)


        # # Get last modifiation date of GCP tflite model
        # storage_client = storage.Client()
        blob = bucket.get_blob(remote_path)
        gcp_model_modified = blob.updated

        _LOGGER.debug(f"GCP model was last modified: {gcp_model_modified}")
        _LOGGER.debug(f"Local model was last modified: {local_model_modified}")

        # If remote model is newer then pull
        if gcp_model_modified > local_model_modified:
            _LOGGER.debug("GCP has a newer version of the model")
            _LOGGER.debug("Pulling model...")
            blob.download_to_filename(local_path)
            _LOGGER.debug("Pull complete!")
        else:
            _LOGGER.debug("local model is up to date.")

    elif remote_exists:
        _LOGGER.debug("Model only exists in remote machine")
        _LOGGER.debug("Pulling model...")
        _LOGGER.debug(os.getcwd())
        blob = bucket.get_blob(remote_path)
        blob.download_to_filename(local_path)
        _LOGGER.debug("Pull complete!")


    elif local_exists:
        _LOGGER.debug("Model only exists in local machine")
        _LOGGER.debug("Model will not be updated")

    else:
        _LOGGER.debug("No model found in Local or Remote storage")

    sense.set_pixels(LED_BLANK)


# load the image, clone it for output, and then convert it to grayscale
def load_img(path):
    if isinstance(path, str):
        image = cv2.imread(path)
    else:
        image = path
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    return image, gray

# find circle (coin) in image then mask and crop
def find_circles(path):
    _LOGGER.debug("start of find_circles()")
    multiple_circles_found = True

    image, gray = load_img(path)
    minDist = 100
    param1 = 40
    param2 = 70
    minRadius = 20
    maxRadius = 1000

    idx = 0 #keep track of itterations in while loop

    while multiple_circles_found:

        # prevent infinite loop if no circle found
        idx = idx + 1
        if idx > 10:
            return 0

        circles = cv2.HoughCircles(gray, cv2.HOUGH_GRADIENT, 1, minDist,
                                   param1=param1, param2=param2, minRadius=minRadius, maxRadius=maxRadius)
        if circles is None:
            param1 = param1 + 5
            param2 = param2 + 5
            continue

        if len(circles[0, :]) == 1:
            multiple_circles_found = False
        elif len(circles[0, :]) > 1:
            param1 = param1 + 5
            param2 = param2 + 5
        else:
            param1 = param1 - 5
            param2 = param2 - 5

    if circles is not None:
        circles = np.uint16(np.around(circles))
        for i in circles[0, :]:
            masked_img = mask_circle(image, i[0], i[1], i[2])
    _LOGGER.debug("end of find_circles()")
    return masked_img

# find difference between two images
# higher MSE -> less similar
def mse(imageA, imageB):
    _LOGGER.debug("start of mse()")
    err = np.sum((imageA.astype("float") - imageB.astype("float")) ** 2)
    err /= float(imageA.shape[0] * imageA.shape[1])
    _LOGGER.debug("end of mse()")
    return err

# mask and crop circle found by find_circles
def mask_circle(image, x, y, r):
    _LOGGER.debug("start of mask_circle()")
    height, width, depth = image.shape
    circle_img = np.zeros((height, width), np.uint8)
    cv2.circle(circle_img, (x, y), r, 1, thickness=-1)

    masked_img = cv2.bitwise_and(image, image, mask=circle_img)

    rectX = (x - r)
    rectY = (y - r)
    masked_img = masked_img[rectY:(rectY + 2 * r), rectX:(rectX + 2 * r)]

    masked_img = cv2.resize(masked_img, (200, 200))
    _LOGGER.debug("end of mask_circle()")
    return masked_img

# save masked and cropped coin image to gcp bucket
def save_to_gcp(label, predict_mode):
    _LOGGER.debug("start of save_to_gcp")
    filename = '/usr/local/foglamp/data/temp.jpg'
    #UPLOADFILE = os.path.join(os.getcwd(), filename)

    if predict_mode:
        gcp_path = f'coins-predicted/{label.lower()}/{time.strftime("%Y%m%d-%H%M%S")}.jpg'
    else:
        gcp_path = f'coins/{label.lower()}/{time.strftime("%Y%m%d-%H%M%S")}.jpg'
    blob = bucket.blob(gcp_path)
    blob.upload_from_filename(filename)
    _LOGGER.debug("end of save_to_gcp()")

# tensor preprocessing for tflite prediction
def set_input_tensor(tflite_interpreter, image):
    _LOGGER.debug("start of set_input_tensor()")
    _LOGGER.debug(tflite_interpreter)
    tensor_index = tflite_interpreter.get_input_details()[0]['index']
    _LOGGER.debug("1")
    input_tensor = tflite_interpreter.tensor(tensor_index)()[0]
    _LOGGER.debug("2")
    input_tensor[:, :] = image
    _LOGGER.debug("end of set_input_tensorr()")

# classify detected coin as heads or tails using tflite model
def classify_image(tflite_interpreter, image, top_k=1):
    _LOGGER.debug("start of classify_image()")
    set_input_tensor(tflite_interpreter, image)
    tflite_interpreter.invoke()
    output_details = tflite_interpreter.get_output_details()[0]
    output = np.squeeze(tflite_interpreter.get_tensor(output_details['index']))
    _LOGGER.debug("end of classify_image()")
    return output





class FrameProcessor(Thread):



    def __init__(self, handle):
        super(FrameProcessor, self).__init__()


        self.heads = 0
        self.tails = 0

        # if floating point model is used we need to subtract the mean and divide
        # by standard deviation
        self.input_mean = handle['input_mean']
        self.input_std = handle['input_std']

        # the height of the detection window on which frames are to be displayed
        self.camera_height = handle['camera_height']

        # the width of the detection window on which frames are to be displayed
        self.camera_width = handle['camera_width']

        model = handle['model_file']['value']
        labels = handle['labels_file']['value']
        self.asset_name = handle['asset_name']['value']
        enable_tpu = handle['enable_edge_tpu']['value']



        global PATH, storage_client, bucket
        PATH = os.path.join(os.getcwd(), handle['gcp_credentials_json']['value'])
        _LOGGER.debug(PATH)
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = PATH
        storage_client = storage.Client(PATH)
        bucket = storage_client.get_bucket(handle['gcp_bucket']['value'])
        _LOGGER.debug(bucket)




        if handle['mode']['value'] == 'predict':
            self.predict = True
            sense.set_pixels(LED_BLANK)
        else:
            self.predict = False
            sense.set_pixels(LED_LABELING)
        self.min_conf_threshold = float(handle['min_conf_threshold']['value'])

        model = os.path.join(os.path.dirname(__file__), "model", model)
        labels = os.path.join(os.path.dirname(__file__), "model", labels)


        _LOGGER.debug(model)
        with open(labels, 'r') as f:
            pairs = (l.strip().split(maxsplit=1) for l in f.readlines())
            labels = dict((int(k), v) for k, v in pairs)


        # instance of the self.inference class
        self.interpreter_loaded = False
        self.inference = Inference()


        interpreter = self.inference.get_interpreter(model, enable_tpu, labels, self.min_conf_threshold)



        if interpreter is not None:
            self.interpreter_loaded = True



        if handle['enable_window']['value'] == 'true' and not FrameProcessor.check_background():
            self.enable_window = True
        else:
            self.enable_window = False

        if handle['source']['value'] == 'camera':
            camera_id = int(handle['camera_id']['value'])
            # Initialize the stream object and start the thread that keeps on reading frames
            # This thread is independent of the Camera Processing Thread
            self.videostream, is_camera_functional = VideoStream(resolution=(self.camera_width,
                                                                             self.camera_height),
                                                                 source=camera_id).start()
            self.is_camera_functional = is_camera_functional

        else:
            stream_url = handle['stream_url']['value']
            stream_protocol = handle['stream_protocol']['value']
            opencv_backend = handle['opencv_backend']['value']
            self.videostream, is_camera_functional = VideoStream(resolution=(self.camera_width,
                                                                             self.camera_height),
                                                                 stream_url=stream_url,
                                                                 stream_protocol=stream_protocol,
                                                                 opencv_backend=opencv_backend).start()
            self.is_camera_functional = is_camera_functional
        # For using the videostream with threading use the following :
        # videostream = VideoStream(resolution=(self.camera_width, self.camera_height),
        # source=source, enable_thread=True).start()
        self.shutdown_in_progress = False






    @staticmethod
    def check_background():
        """
        Checks for foglamp running in background.
        Returns: True if foglamp is running in background.

        """
        out = subprocess.Popen(['systemctl', 'status', 'foglamp'],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
        stdout, stderr = out.communicate()
        if str(stdout).find("active (running)") != -1:
            return True
        else:
            return False

    @staticmethod
    def construct_readings(objs, predict_mode):
        _LOGGER.debug("start of construct_readings()")
        reads = {}

        if len(objs) == 0:
            return 0
        data = {}
        _LOGGER.debug(objs)

        if predict_mode:

            for r_index in range(len(objs)):
                if objs[r_index]['label'] == "Heads":
                    reads['Heads'] = 1
                    reads['Tails'] = 0
                    reads['confidence'] = objs[r_index]['score']
                elif objs[r_index]['label'] == "Tails":
                    reads['Heads'] = 0
                    reads['Tails'] = 1
                    reads['confidence'] = objs[r_index]['score']
                else:
                    reads['Heads'] = 0
                    reads['Tails'] = 0
                    reads['confidence'] = 0


            data = {
                'asset': "coin_detection_predictions",
                'timestamp': utils.local_timestamp(),
                'readings': reads
            }
        else:
            for r_index in range(len(objs)):
                reads['heads_recorded'] = objs[r_index]['heads_recorded']
                reads['tails_recorded'] = objs[r_index]['tails_recorded']
            data = {
                'asset': "coin_detection_collection",
                'timestamp': utils.local_timestamp(),
                'readings': reads
            }


        _LOGGER.debug("end of construct_readings()")
        return data

    def wait_for_frame(self):
        """ Waits for frame to become available else sleeps for 200 milliseconds.
                Args:
                       self-> a videostream object
                Returns: None
               Raises: None
           """
        while True:
            if self.videostream.frame is not None:
                return
            else:
                time.sleep(0.2)

    def handle_new_config(self, new_config):
        """
        If shutdown is not required then it changes the configuration on the fly.
        Args:
            new_config: The configuration during reconfigure.

        Returns:
              None
        """

        _LOGGER.debug("Handling the reconfigure without shutdown")



        global PATH, storage_client, bucket
        PATH = os.path.join(os.getcwd(), new_config['gcp_credentials_json']['value'])
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = PATH
        storage_client = storage.Client(PATH)
        bucket = storage_client.get_bucket(new_config['gcp_bucket']['value'])


        pull_model()

        model = new_config['model_file']['value']
        labels = new_config['labels_file']['value']

        self.asset_name = new_config['asset_name']['value']

        enable_tpu = new_config['enable_edge_tpu']['value']

        if new_config['mode']['value'] == "predict":
            self.predict = True
            sense.set_pixels(LED_BLANK)
        else:
            self.predict = False
            sense.set_pixels(LED_LABELING)

        self.min_conf_threshold = float(new_config['min_conf_threshold']['value'])

        model = os.path.join(os.path.dirname(__file__), "model", model)
        labels = os.path.join(os.path.dirname(__file__), "model", labels)


        with open(labels, 'r') as f:
            pairs = (l.strip().split(maxsplit=1) for l in f.readlines())
            labels = dict((int(k), v) for k, v in pairs)

        _ = self.inference.get_interpreter(model, enable_tpu, labels, self.min_conf_threshold)


        _LOGGER.debug("Handled the reconfigure")



    def run(self):
        mode = "predict" if self.predict else "collect"
        _LOGGER.debug("Beginning main loop in %s mode", mode)

        # The thread is allowed to capture a few frames. See FOGL-4132 for details
        self.wait_for_frame()

        prev_frame = None
        while True:
            global c_callback, c_ingest_ref
            """
            # we need the height , width to resize the image for feeding into the model
            height_for_model = self.inference.height_for_model
            width_for_model = self.inference.width_for_model
            """
            #  check if floating point model is used or not
            floating_model = self.inference.floating_model

            # The list of labels of the supported objects detected by the plugin
            labels = self.inference.labels


            # Taking the frame the stream
            frame = self.videostream.read()
            if frame is None:
                _LOGGER.error("Either the stream/camera device stopped working.")
                break

            objs = []
            #_LOGGER.debug('Predict bool: %s', str(self.predict))
            if self.predict:
                #_LOGGER.debug("Inside inference mode")
                masked_img = find_circles(frame)

                if isinstance(masked_img, int):
                    _LOGGER.debug("No coin found in image")
                    sense.set_pixels(LED_BLANK)
                    predicted_class = "None"
                    predicted_class_int = 2



                elif prev_frame is not None and mse(frame, prev_frame) < 200:
                    _LOGGER.debug("Current frame too similar to previously forwarded image")
                else:
                    _LOGGER.debug("Saving detected coin to temp image")
                    prev_frame = frame
                    cv2.imwrite("/usr/local/foglamp/data/temp.jpg", masked_img)

                    if floating_model:
                        masked_img = (np.float32(masked_img / 255.0))

                    pred_conf = self.inference.perform_inference(masked_img)
                    pred = pred_conf > float(self.min_conf_threshold)
                    if pred:
                        predicted_class = "Tails"
                        predicted_class_int = 1
                        sense.set_pixels(LED_TAILS)

                    else:
                        predicted_class = "Heads"
                        predicted_class_int = 0
                        sense.set_pixels(LED_HEADS)

                    _LOGGER.debug("Predicted Side Up: %s    With Confidence: %s", predicted_class, pred_conf)
                    save_to_gcp(predicted_class, self.predict)

                objs.append({'label': predicted_class, 'label_int': predicted_class_int, 'score': float(pred_conf)})
            else:
                event = sense.stick.get_events()
                if len(event) > 0 and event[-1].action == "pressed":
                    masked_img = find_circles(frame)

                    if isinstance(masked_img, int):
                        _LOGGER.info("No coin detected in frame...")
                        _LOGGER.info("Not adding image to dataset.")
                        sense.set_pixels(LED_NO_COIN)
                    elif prev_frame is not None and mse(frame, prev_frame) < 200:
                        _LOGGER.info("Image to similar to last image forwarded to GCP...")
                        _LOGGER.info("Not adding image to dataset.")
                        sense.set_pixels(LED_MOVE_COIN)
                    else:
                        prev_frame = frame
                        cv2.imwrite("/usr/local/foglamp/data/temp.jpg", masked_img)
                        if event[-1].direction == "down":
                            recorded_label = "Heads"
                            self.heads += 1
                            sense.set_pixels(LED_HEADS_RECORDED)
                        else:
                            recorded_label = "Tails"
                            self.tails += 1
                            sense.set_pixels(LED_TAILS_RECORDED)

                        _LOGGER.info("Adding %s to GCP dataset", recorded_label)
                        save_to_gcp(recorded_label, self.predict)
                        _LOGGER.debug("heads %s tails %s", self.heads, self.tails)
                        objs.append({'heads_recorded': self.heads, 'tails_recorded': self.tails})


            # All the results have been drawn on the frame, so it's time to display it.
            if self.shutdown_in_progress:
                _LOGGER.debug("Shut down breaking loop")
                break
            else:
                if len(objs) != 0:
                    data = FrameProcessor.construct_readings(objs, self.predict)
                    if data:
                        async_ingest.ingest_callback(c_callback, c_ingest_ref, data)

            try:
                if self.enable_window:
                    cv2.imshow("Coin Detector", frame)

                WebStream.FRAME = frame.copy()

            except Exception as e:
                _LOGGER.info('exception  {}'.format(e))

            # wait for 1 milli second
            cv2.waitKey(1)

        WebStream.SHUTDOWN_IN_PROGRESS = True
        _LOGGER.debug("Shutdown flag of streaming server set True")
        _LOGGER.debug("Stopping the stream ")
        self.videostream.stop()
        _LOGGER.info("Camera stream has been stopped")
        time.sleep(2)
        cv2.destroyAllWindows()
        _LOGGER.debug("All windows destroyed")
