# foglamp_END

""" Helper Class for loading the model and performing the inference on it
"""

import logging
import os

import numpy as np

from foglamp.common import logger

_LOGGER = logger.setup(__name__, level=logging.DEBUG)

EDGETPU_SHARED_LIB = 'libedgetpu.so.1'


try:
    from tflite_runtime.interpreter import Interpreter
    try:
        from tflite_runtime.interpreter import load_delegate
    except ImportError:
        _LOGGER.debug("Edge TPU support is not found in your tensorflow installation!")
        pass
except ImportError as e:
    _LOGGER.debug("Tensorflow installation not found.")


class Inference:
    def __init__(self):
        _LOGGER.debug("creating inference class")
        self.interpreter = None
        self.input_details = None
        self.output_details = None
        self.height_for_model = None
        self.width_for_model = None
        self.floating_model = None
        self.min_conf_threshold = None
        self.labels = None

    def get_interpreter(self, model, enable_tpu, labels, min_conf_threshold=0.5):
        """ Returns interpreter from the model
            Args:
                  model -> full path of the .tflite file ,
                  enable_tpu-> whether you want to use Edge TPU or not
        Returns:
                  interpreter object
        Raises:
                Raises exception if the Runtime Library for Edge TPU
                is not found

        """

        _LOGGER.debug("inside of get interpreter")

        interpreter_loaded = False
        if enable_tpu == 'true':
            try:
                # loading the Edge TPU Runtime
                model, *device = model.split('@')
                if os.path.exists(model):
                    load_delegate(EDGETPU_SHARED_LIB)
                    self.interpreter = Interpreter(model_path=model,
                                                   experimental_delegates=[load_delegate(EDGETPU_SHARED_LIB,
                                                                         {'device': device[0]} if device else {})])
                    interpreter_loaded = True

                else:
                    _LOGGER.exception("Please make sure the model file exists")

            except OSError:
                 _LOGGER.exception("Please install runtime for edge tpu")

            except ValueError:
                _LOGGER.exception("Make sure edge tpu is plugged in")

        else:
            _LOGGER.debug(model)
            if os.path.exists(model):
                _LOGGER.debug(os.getcwd())
                self.interpreter = Interpreter(model_path="../python/foglamp/plugins/south/coin_detection/model/coin_model.tflite")
                self.interpreter.allocate_tensors()
                interpreter_loaded = True
            else:
                _LOGGER.exception("Please make sure the model file exists")

        if interpreter_loaded:
            _LOGGER.debug("Interpreter Loaded")
            self.interpreter.allocate_tensors()
            self.input_details = self.interpreter.get_input_details()
            _LOGGER.debug(self.input_details)
            self.output_details = self.interpreter.get_output_details()
            self.height_for_model = self.input_details[0]['shape'][1]
            self.width_for_model = self.input_details[0]['shape'][2]
            self.floating_model = (self.input_details[0]['dtype'] == np.float32)
            self.min_conf_threshold = min_conf_threshold
            self.labels = labels
        else:
            _LOGGER.error("The interpreter requested is not loaded. Try again after shutting down.")
            return None

        # return None if not loaded
        return self.interpreter

    def perform_inference(self, input_data):
        """ Returns bounding box , class , score
            Args:
                   input_data ->  the input data to be fed into the model
            Returns:
                     scores-> an array of scores(0 to 1) of the objects detected
            Raises: None
        """


        _LOGGER.debug("start of perform_inference()")

        input_data = np.expand_dims(input_data, axis=0)
        _LOGGER.debug(self.input_details[0]['index'])
        _LOGGER.debug(input_data)
        self.interpreter.set_tensor(self.input_details[0]['index'], input_data)

        self.interpreter.invoke()
        # Retrieve detection results
        _LOGGER.debug(self.interpreter.get_output_details()[0])
        output_details = self.interpreter.get_output_details()[0]
        output = np.squeeze(self.interpreter.get_tensor(output_details['index']))
        _LOGGER.debug("end of perform_inference()")
        return output
 
